@@ -1,14 +1,14 @@
 package flowcontrol
 
 import (
-	"github.com/golang/mock/gomock"
-	. "github.com/onsi/ginkgo"
-	. "github.com/onsi/gomega"
-
 	"testing"
+
+	"github.com/golang/mock/gomock"
+	. "github.com/onsi/ginkgo/v2"
+	. "github.com/onsi/gomega"
 )
 
-func TestCrypto(t *testing.T) {
+func TestFlowControl(t *testing.T) {
 	RegisterFailHandler(Fail)
 	RunSpecs(t, "FlowControl Suite")
 }
